from eincheck.parser.grammar import parse_shape_spec

parser_cache_info = parse_shape_spec.cache_info
parser_cache_clear = parse_shape_spec.cache_clear
parser_resize_cache = parse_shape_spec.reset_maxsize



from __future__ import annotations

import dataclasses
import enum
from typing import Collection, Dict, Optional, Union

from eincheck.parser.expressions import DataExpr, Expr, Literal, Variable
from eincheck.types import ShapeVariable


@enum.unique
class DimType(enum.Enum):
    # Expression evaluates to a single integer, DimSpec matches a single dimension.
    SINGLE = "single"
    # Expression evaluates to a tuple, DimSpec matches multiple dimensions.
    VARIADIC = "variadic"
    # Expression evaluates to a single integer, DimSpec matches multiple dimensions.
    REPEATED = "repeated"


@dataclasses.dataclass(frozen=True)
class DimSpec:
    """A specification for one part of a shape.

    This can be more than one dimension if matches_multiple is True.
    A full shape spec is made up of a sequence of DimSpecs.
    """

    # Expression for the expected value of this DimSpec.
    # This can evaluate to a single int (if not variadic) or a
    # tuple of ints (if variadic).
    # None represents an unconstrained dimension (i.e. _ or ...).
    value: Optional[Expr]
    type: DimType = dataclasses.field(default=DimType.SINGLE)
    # If true, will accept any shape that can be broadcast to value.
    can_broadcast: bool = dataclasses.field(default=False)

    def __str__(self) -> str:
        s = str(self.value) if self.value else "_"
        if self.type is DimType.VARIADIC:
            s = "*" + s
        elif self.type is DimType.REPEATED:
            s = s + "*"
        if self.can_broadcast:
            s = s + "!"
        return s

    @property
    def matches_multiple(self) -> bool:
        return self.type is not DimType.SINGLE

    def make_variadic(self) -> DimSpec:
        assert self.type is DimType.SINGLE
        return dataclasses.replace(self, type=DimType.VARIADIC)

    def make_repeated(self) -> DimSpec:
        assert self.type is DimType.SINGLE
        return dataclasses.replace(self, type=DimType.REPEATED)

    def make_can_broadcast(self) -> DimSpec:
        return dataclasses.replace(self, can_broadcast=True)

    @staticmethod
    def create_literal(x: int) -> DimSpec:
        return DimSpec(Literal(x))

    @staticmethod
    def create_variable(x: str) -> DimSpec:
        return DimSpec(Variable(x))

    @staticmethod
    def create(x: Union[DimSpec, int, str, None]) -> DimSpec:
        if isinstance(x, DimSpec):
            return x
        elif x == "$":
            return DimSpec(DataExpr())
        elif x == "_":
            return DimSpec(None)
        elif isinstance(x, str) and x.isdigit():
            return DimSpec.create_literal(int(x))
        elif isinstance(x, str):
            return DimSpec.create_variable(x)
        elif isinstance(x, int):
            return DimSpec.create_literal(x)
        elif x is None:
            return DimSpec(None)
        else:
            raise ValueError(f"Unexpected type: {type(x).__name__}")

    def is_defined(self, values: Collection[str]) -> bool:
        """Whether self.value can be evaluated given a set of known variables."""
        if self.value is None:
            return True
        return self.value.is_defined(values)

    def is_checkable(self, values: Collection[str]) -> bool:
        """Whether the DimSpec can be used in a shape check with the given values."""
        return (
            isinstance(self.value, Variable) and not self.can_broadcast
        ) or self.is_defined(values)

    def n_dims(self, bindings: Dict[str, ShapeVariable]) -> Optional[int]:
        """Determine the number of dimensions matched by this DimSpec."""
        if self.type is DimType.SINGLE:
            return 1
        if self.type is DimType.REPEATED:
            return None

        if self.value is not None and self.is_defined(bindings):
            got = self.value.eval(bindings)
            if isinstance(got, tuple):
                return len(got)

        return None



from contextlib import contextmanager
from typing import ContextManager, Generator

__EINCHECK_ENABLE_CHECKS: bool = True


def _should_do_checks() -> bool:
    return __EINCHECK_ENABLE_CHECKS


@contextmanager
def _set_enable_checks(value: bool) -> Generator[None, None, None]:
    global __EINCHECK_ENABLE_CHECKS
    prev = __EINCHECK_ENABLE_CHECKS
    __EINCHECK_ENABLE_CHECKS = value
    yield
    assert __EINCHECK_ENABLE_CHECKS == value
    __EINCHECK_ENABLE_CHECKS = prev


def enable_checks() -> ContextManager[None]:
    """Enable eincheck to do shape checks."""
    return _set_enable_checks(True)


def disable_checks() -> ContextManager[None]:
    """Disable eincheck from doing shape checks."""
    return _set_enable_checks(False)



from typing import Any, List, Optional, Sequence, Tuple

from eincheck.types import Tensor


def get_shape(x: Tensor) -> Optional[Tuple[Optional[int], ...]]:
    if hasattr(x, "shape"):
        s: Optional[Sequence[Optional[int]]] = x.shape
        if s is not None:
            return tuple(i if isinstance(i, int) else None for i in s)

    elif isinstance(x, (tuple, list)):
        return (len(x),)

    return None


def parse_dot_name(dot_name: str) -> Tuple[str, List[str]]:
    """Split a dot-separated name path into pieces."""
    first_name, *name_path = dot_name.split(".")
    return first_name, name_path


def get_object(dot_name: str, data: Any) -> Any:
    """Use a dot name path to extract an object from data."""
    if not dot_name:
        return data

    first_name, name_path = parse_dot_name(dot_name)

    x = _get_field(data, first_name)
    for n in name_path:
        x = _get_field(x, n)

    return x


def _get_field(obj: Any, field: str) -> Any:
    if obj is None:
        return None

    if field.isdecimal():
        idx = int(field)
        return obj[idx]

    if isinstance(obj, dict):
        return obj[field]

    return getattr(obj, field)



import itertools
from typing import Any, Dict, Optional, Set, Tuple, Union, cast

from eincheck.contexts import _should_do_checks
from eincheck.parser.dim_spec import DimSpec, DimType
from eincheck.parser.expressions import DataExpr, Variable
from eincheck.parser.grammar import ShapeArg, create_shape_spec
from eincheck.parser.shape_spec import ShapeSpec
from eincheck.types import ShapeVariable, Tensor
from eincheck.utils import get_shape


def _check_dim_spec(
    got_shape: Tuple[int, ...],
    d: DimSpec,
    bindings: Dict[str, ShapeVariable],
    name: str,
    msg: str,
    start_idx: int,
) -> None:
    if d.value is None:
        return

    expected_value = d.value.eval(bindings)

    if d.can_broadcast:
        broadcast_values: Set[ShapeVariable]
        if isinstance(expected_value, int):
            broadcast_values = {expected_value, 1}
        else:
            broadcast_values = {
                tuple(p) for p in itertools.product(*([x, 1] for x in expected_value))
            }

    if d.type is DimType.VARIADIC and isinstance(expected_value, int):
        raise ValueError(
            f"{name}: expected variadic DimSpec {d} to evaluate to a tuple, "
            f"got {expected_value}{msg}"
        )
    elif d.type is not DimType.VARIADIC and isinstance(expected_value, tuple):
        raise ValueError(
            f"{name}: expected non-variadic DimSpec {d} to evaluate to an integer, "
            f"got {expected_value}{msg}"
        )

    def do_check(g: ShapeVariable, indices: ShapeVariable) -> None:
        if d.can_broadcast and g not in broadcast_values:
            first_line = f"expected can broadcast to {d.value}={expected_value} got {g}"
        elif not d.can_broadcast and g != expected_value:
            first_line = f"expected {d.value}={expected_value} got {g}"
        else:
            first_line = None

        if first_line:
            dim_str = (
                f"dim {indices}" if isinstance(indices, int) else f"dims {indices}"
            )
            raise ValueError(
                f"{name} {dim_str}: {first_line}"
                + "".join(f"\n    {k}={v}" for k, v in bindings.items())
                + msg
            )

    if d.type is DimType.REPEATED:
        for g_idx, g in enumerate(got_shape):
            do_check(g, start_idx + g_idx)
    elif d.type is DimType.VARIADIC:
        do_check(got_shape, tuple(range(start_idx, start_idx + len(got_shape))))
    elif len(got_shape) != 1:
        raise RuntimeError(
            f"{name}: expected a single dimension for {d}, got {got_shape}{msg}"
        )
    else:
        do_check(got_shape[0], start_idx)


def _bind_shape(
    got_shape: Tuple[Optional[int], ...],
    s: ShapeArg,
    bindings: Dict[str, ShapeVariable],
    name: str,
    msg: str,
) -> None:
    expected_shape = create_shape_spec(s)

    for d, start_idx, end_idx in expected_shape.matched_indices(
        bindings, len(got_shape)
    ):
        if not isinstance(d.value, Variable) or d.value.x in bindings:
            continue

        g_slice = got_shape[start_idx:end_idx]
        if None in g_slice:
            raise ValueError(
                f"{name}: tried to match {g_slice} to {d}, found None{msg}"
            )

        g_slice = cast(Tuple[int, ...], g_slice)

        if d.type is DimType.VARIADIC:
            bindings.setdefault(d.value.x, g_slice)
        elif d.type is DimType.REPEATED and g_slice:
            bindings.setdefault(d.value.x, g_slice[0])
        elif d.type is DimType.SINGLE and len(g_slice) != 1:
            raise RuntimeError(
                f"{name}: expected a single dimension for {d}, got {got_shape}{msg}"
            )
        elif d.type is DimType.SINGLE:
            bindings.setdefault(d.value.x, g_slice[0])
        else:
            # Only reach here if d.type is REPEATED and not got_shape.
            # Nothing to check in this case.
            pass


def _check_shape(
    got_shape: Tuple[Optional[int], ...],
    s: ShapeArg,
    bindings: Dict[str, ShapeVariable],
    name: str,
    msg: str,
) -> Dict[str, ShapeVariable]:
    expected_shape = create_shape_spec(s)

    if msg:
        msg = "\n" + msg

    _check_rank(name, got_shape, expected_shape, bindings, msg)

    unknown_size_inds = expected_shape.unknown_n_dims_indices(bindings)
    if len(unknown_size_inds) > 1:
        raise RuntimeError(
            f"{name} has multiple DimSpec of unknown size: "
            f"{[expected_shape.dims[i] for i in unknown_size_inds]}" + msg
        )

    for d, start_idx, end_idx in expected_shape.matched_indices(
        bindings, len(got_shape)
    ):
        if d.value is None:
            continue

        if d.type is DimType.REPEATED and start_idx == end_idx:
            continue

        g_slice = got_shape[start_idx:end_idx]
        if None in g_slice:
            inds = (
                f"dim {start_idx}"
                if end_idx == start_idx + 1
                else f"dims {tuple(range(start_idx, end_idx))}"
            )
            raise ValueError(
                f"{name} {inds}: tried to check {d} against {g_slice}, found None{msg}"
            )

        _check_dim_spec(
            cast(Tuple[int, ...], g_slice), d, bindings, name, msg, start_idx
        )

    return bindings


def check_shapes(
    *args: Tuple[Tensor, ShapeArg],
    **kwargs: Union[ShapeVariable, Tuple[Tensor, ShapeArg]],
) -> Dict[str, ShapeVariable]:
    """Check the shapes of Tensors against ShapeArg specifications.

    Examples:

    .. doctest::

        >>> from numpy.random import randn
        >>> from eincheck import check_shapes
        >>>
        >>> check_shapes((randn(3, 4, 5), "... i j"), (randn(5, 6), "... j k"))
        {'i': 4, 'j': 5, 'k': 6}
        >>> check_shapes(
        ...     x=(randn(8, 2, 7, 3), "*batch t 3"),
        ...     y=(randn(8, 2, 1, 1, 3), "*batch ... 3"),
        ...     batch=(8, 2),
        ... )
        {'batch': (8, 2), 't': 7}

    Pass pairs of (tensor, shape spec) as either args or kwargs. The only difference
    when using a kwarg is the display name in error messages (args are called ``arg0``,
    ``arg1``, etc).

    Kwargs can also specify the variable values (e.g. ``batch=(8, 2)``).

    If all shape checks pass, returns a dictionary with all variable values.

    :param args: Pairs of (tensor, shape spec)
    :param kwargs: Pairs of (tensor, shape spec)
    :raise ValueError: If the shapes are incorrect or cannot be verified
    :return: Values for all bound variables from the shape specs
    """
    if not _should_do_checks():
        return {}

    tensors, bindings = _get_tensors_and_bindings(*args, **kwargs)
    _check_variable_types(tensors, bindings)

    if not tensors:
        return bindings

    got_msgs = [str(x) for x, _ in tensors.values()]
    got_len = max(map(len, got_msgs))
    msg = "\n".join(
        f"  {name}: got {g:<{got_len}} expected {s}"
        for (name, (_, s)), g in zip(tensors.items(), got_msgs)
    )

    checked_names = set()
    binded_names = set()

    for _ in range(len(tensors)):
        bindings_len = len(bindings)

        for t_name, (t_got, t_expected) in tensors.items():
            if (
                t_name in checked_names
                or len(t_expected.unknown_n_dims_indices(bindings)) > 1
            ):
                continue

            if t_name not in binded_names:
                _check_rank(t_name, t_got, t_expected, bindings, msg)
                _bind_shape(t_got, t_expected, bindings, t_name, msg)
                binded_names.add(t_name)

            if not t_expected.is_checkable(bindings):
                continue

            _check_shape(t_got, t_expected, bindings, t_name, msg)
            checked_names.add(t_name)

        if len(bindings) == bindings_len:
            break

    unbound = set(tensors) - binded_names
    if unbound:
        raise ValueError(
            f"Unable to determine bindings for: {' '.join(unbound)}\n{msg}"
        )
    unchecked = set(tensors) - checked_names
    if unchecked:
        missing_vars = set.union(*(tensors[k][1].variables for k in unchecked))
        raise ValueError(
            f"Unable to check: [{' '.join(sorted(unchecked))}] "
            f"missing variables: [{' '.join(sorted(missing_vars))}]\n{msg}"
        )

    return bindings


def _check_variable_types(
    tensors: Dict[str, Tuple[Tuple[Optional[int], ...], ShapeSpec]],
    bindings: Dict[str, ShapeVariable],
) -> None:
    """Check that each variable is either an int or a tuple.

    Categorizes each variable present in the ShapeSpec and bindings as being either an
    int or a tuple.

    If any variables are in both sets, raises an error.
    """

    int_vars = set()
    tuple_vars = set()

    for k, v in bindings.items():
        if isinstance(v, int):
            int_vars.add(k)
        else:
            tuple_vars.add(k)

    for _, spec in tensors.values():
        for dim in spec.dims:
            if not dim.value:
                continue

            if dim.type is DimType.VARIADIC:
                tuple_vars.update(dim.value.variables)
            else:
                int_vars.update(dim.value.variables)

    both = int_vars & tuple_vars
    if both:
        raise ValueError(
            "Found variables in both variadic and non-variadic expressions: "
            + " ".join(sorted(both))
        )


def _check_rank(
    name: str,
    got_shape: Tuple[Optional[int], ...],
    shape_spec: ShapeSpec,
    bindings: Dict[str, ShapeVariable],
    msg: str,
) -> None:
    expected_rank = shape_spec.min_rank(bindings)

    if shape_spec.unknown_n_dims_indices(bindings):
        bound_text = "at least "
        check = len(got_shape) < expected_rank
    else:
        bound_text = ""
        check = len(got_shape) != expected_rank

    if check:
        rows = [
            f"{name}: expected rank {bound_text}{expected_rank}, "
            f"got shape {got_shape}"
        ] + [f"  {k} = {v}" for k, v in bindings.items()]
        raise ValueError("\n".join(rows) + "\n" + msg)


def _get_tensors_and_bindings(
    *args: Tuple[Tensor, ShapeArg],
    **kwargs: Union[ShapeVariable, Tuple[Tensor, ShapeArg]],
) -> Tuple[
    Dict[str, Tuple[Tuple[Optional[int], ...], ShapeSpec]], Dict[str, ShapeVariable]
]:
    tensors: Dict[str, Tuple[Tuple[Optional[int], ...], ShapeSpec]] = {}

    for idx, (a_tensor, a_shape) in enumerate(args):
        tensors.update(_get_shapes(a_tensor, create_shape_spec(a_shape), f"arg{idx}"))

    bindings: Dict[str, ShapeVariable] = {}

    for k, v in kwargs.items():
        if isinstance(v, int) or (
            isinstance(v, tuple) and all(isinstance(vv, int) for vv in v)
        ):
            bindings[k] = cast(ShapeVariable, v)
        elif isinstance(v, tuple) and len(v) == 2:
            v = cast(Tuple[Tensor, ShapeArg], v)
            assert k not in tensors
            tensors.update(_get_shapes(v[0], create_shape_spec(v[1]), k))
        else:
            raise ValueError(f"Unexpected kwarg {v}")

    return tensors, bindings


def _get_shapes(
    x: Any, s: ShapeSpec, name: str
) -> Dict[str, Tuple[Tuple[Optional[int], ...], ShapeSpec]]:
    if s.is_data_expr:
        if not hasattr(x, "_get_shapes"):
            raise ValueError(
                f"{name}: spec $ specified, but no _get_shapes method was found. "
                "This should have been added by the @check_data decorator."
            )

        return {
            k2: v2
            for k, (vt, vs) in x._get_shapes().items()
            for k2, v2 in _get_shapes(vt, vs, f"{name}.{k}").items()
        }

    shape = get_shape(x)
    if shape is not None:
        if any(isinstance(d.value, DataExpr) for d in s.dims):
            raise ValueError(
                f"{name}: $ should not be present in the shape spec for a Tensor, "
                f"got {s}"
            )
        return {name: (shape, s)}

    return {}



from typing import Optional, Sequence, Tuple, TypeVar, Union

_ShapeType = TypeVar("_ShapeType", bound=Optional[Sequence[Optional[int]]])


Tensor = TypeVar("Tensor")
ShapeVariable = Union[int, Tuple[int, ...]]



from typing import Any, Sequence, Tuple, Union

from lark.lark import Lark
from lark.visitors import Transformer

from eincheck.cache import resizeable_lru_cache
from eincheck.parser.dim_spec import DimSpec, DimType
from eincheck.parser.expressions import (
    AddOp,
    BroadcastOp,
    ConcatOp,
    DataExpr,
    Expr,
    MulOp,
    SubOp,
)
from eincheck.parser.shape_spec import ShapeSpec

ShapeArg = Union[str, ShapeSpec, Sequence[Union[DimSpec, str, int, None]]]

grammar = r"""

shape : can_broadcast_dim? (" "+ can_broadcast_dim)*
      | "$" -> dollar

?expr : value_expr
      | "_" -> underscore

?value_expr : INT   -> int
            | NAME  -> word
            | math

?can_broadcast_dim : dim "!" -> can_broadcast
                   | dim

?dim : expr "*" -> repeated
     | "*" expr -> variadic
     | expr
     | "..." -> ellipse

?math : "(" " "* value_expr " "* "+" " "* value_expr " "* ")" -> add
      | "(" " "* value_expr " "* "-" " "* value_expr " "* ")" -> sub
      | "(" " "* value_expr " "* "*" " "* value_expr " "* ")" -> mul
      | "(" " "* value_expr " "* "||" " "* value_expr " "* ")" -> concat
      | "(" " "* value_expr " "* "^" " "* value_expr " "* ")" -> broadcast
      | "(" " "* value_expr " "* ")"

%import common.INT
%import python.NAME
"""

_parser = Lark(grammar, start="shape")


class TreeToSpec(Transformer):  # type: ignore[type-arg]
    def int(self, s: Any) -> DimSpec:
        (x,) = s
        return DimSpec.create_literal(int(x))

    def word(self, s: Any) -> DimSpec:
        (x,) = s
        return DimSpec.create_variable(x)

    def add(self, s: Any) -> DimSpec:
        x, y = s
        return DimSpec(AddOp(self._get_expr(x), self._get_expr(y)))

    def mul(self, s: Any) -> DimSpec:
        x, y = s
        return DimSpec(MulOp(self._get_expr(x), self._get_expr(y)))

    def sub(self, s: Any) -> DimSpec:
        x, y = s
        return DimSpec(SubOp(self._get_expr(x), self._get_expr(y)))

    def concat(self, s: Any) -> DimSpec:
        x, y = s
        return DimSpec(ConcatOp(self._get_expr(x), self._get_expr(y)))

    def broadcast(self, s: Any) -> DimSpec:
        x, y = s
        return DimSpec(BroadcastOp(self._get_expr(x), self._get_expr(y)))

    def underscore(self, s: Any) -> DimSpec:
        return DimSpec(None)

    def repeated(self, s: Tuple[DimSpec]) -> DimSpec:
        (x,) = s
        assert isinstance(x, DimSpec)
        return x.make_repeated()

    def variadic(self, s: Tuple[DimSpec]) -> DimSpec:
        (x,) = s
        assert isinstance(x, DimSpec)
        return x.make_variadic()

    def can_broadcast(self, s: Tuple[DimSpec]) -> DimSpec:
        (x,) = s
        assert isinstance(x, DimSpec)
        return x.make_can_broadcast()

    def ellipse(self, s: Any) -> DimSpec:
        return DimSpec(None, DimType.REPEATED)

    def shape(self, dims: Sequence[DimSpec]) -> ShapeSpec:
        return ShapeSpec(list(dims))

    def dollar(self, s: Any) -> ShapeSpec:
        return ShapeSpec([DimSpec(DataExpr())])

    @staticmethod
    def _get_expr(x: Any) -> Expr:
        assert isinstance(x, DimSpec)
        assert not x.matches_multiple
        assert x.value is not None
        return x.value


_transformer = TreeToSpec()


@resizeable_lru_cache()
def parse_shape_spec(s: str) -> ShapeSpec:
    """Parse a string into a ShapeSpec."""
    out = _transformer.transform(_parser.parse(s.strip(" ")))
    assert isinstance(out, ShapeSpec)
    return out


def create_shape_spec(arg: ShapeArg) -> ShapeSpec:
    """Convert a generic ShapeArg into a ShapeSpec."""
    if isinstance(arg, ShapeSpec):
        return arg
    if isinstance(arg, str):
        return parse_shape_spec(arg)

    dims = [DimSpec.create(x) for x in arg]
    return ShapeSpec(dims)



import operator
import re
from abc import ABC, abstractmethod
from typing import Any, Callable, Collection, Dict, Generic, Set, Tuple, TypeVar

from eincheck.types import ShapeVariable


class Expr(ABC):
    """Base expression."""

    @abstractmethod
    def __str__(self) -> str:
        """Representation of the expression."""

    @abstractmethod
    def eval(self, values: Dict[str, ShapeVariable]) -> ShapeVariable:
        """Evaluate the expression given existing values."""

    @property
    @abstractmethod
    def variables(self) -> Set[str]:
        """The set of variables used for this expression."""

    def is_defined(self, values: Collection[str]) -> bool:
        """Whether all the necessary variables are defined."""
        return self.variables <= set(values)


class Literal(Expr):
    """A literal value."""

    def __init__(self, x: int):
        self.x = x

    def __str__(self) -> str:
        return str(self.x)

    def eval(self, values: Dict[str, ShapeVariable]) -> ShapeVariable:
        return self.x

    @property
    def variables(self) -> Set[str]:
        return set()

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Literal) and o.x == self.x


class Variable(Expr):
    NAME_REGEX = re.compile(r"[^\W\d]\w*")

    def __init__(self, x: str):
        if not x:
            raise ValueError("Variable name must not be empty")
        if not self.NAME_REGEX.fullmatch(x):
            raise ValueError(f"Variable name should be a valid python name, got {x}")
        self.x = str(x)

    def __str__(self) -> str:
        return self.x

    def eval(self, values: Dict[str, ShapeVariable]) -> ShapeVariable:
        return values[self.x]

    @property
    def variables(self) -> Set[str]:
        return {self.x}

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Variable) and o.x == self.x


_T = TypeVar("_T", int, Tuple[int, ...])


class _BinaryOp(Expr, Generic[_T]):
    expected_type: Any

    def __init__(self, x: Expr, y: Expr, op: str, func: Callable[[_T, _T], _T]):
        self.x = x
        self.y = y
        self.op = op
        self.func: Callable[[_T, _T], _T] = func

    def __str__(self) -> str:
        return f"({self.x}{self.op}{self.y})"

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, type(self))
            and o.x == self.x
            and o.y == self.y
            and o.op == self.op
        )

    def eval(self, values: Dict[str, ShapeVariable]) -> _T:
        x = self.x.eval(values)
        y = self.y.eval(values)
        if not isinstance(x, self.expected_type):
            raise ValueError(
                f"Expected {self.expected_type.__name__} for {self.x} in {self}, "
                f"got {x}"
            )
        if not isinstance(y, self.expected_type):
            raise ValueError(
                f"Expected {self.expected_type.__name__} for {self.y} in {self}, "
                f"got {y}"
            )

        return self.func(x, y)

    @property
    def variables(self) -> Set[str]:
        return self.x.variables | self.y.variables


class _ScalarBinaryOp(_BinaryOp[int]):
    expected_type = int


class AddOp(_ScalarBinaryOp):
    def __init__(self, x: Expr, y: Expr):
        super().__init__(x, y, "+", operator.add)


class MulOp(_ScalarBinaryOp):
    def __init__(self, x: Expr, y: Expr):
        super().__init__(x, y, "*", operator.mul)


class SubOp(_ScalarBinaryOp):
    def __init__(self, x: Expr, y: Expr):
        super().__init__(x, y, "-", operator.sub)


class _TupleBinaryOp(_BinaryOp[Tuple[int, ...]]):
    expected_type = tuple


class ConcatOp(_TupleBinaryOp):
    def __init__(self, x: Expr, y: Expr):
        super().__init__(x, y, "||", operator.add)


class BroadcastOp(_TupleBinaryOp):
    def __init__(self, x: Expr, y: Expr):
        super().__init__(x, y, "^", BroadcastOp.broadcast)

    @staticmethod
    def broadcast(x: Tuple[int, ...], y: Tuple[int, ...]) -> Tuple[int, ...]:
        len_diff = len(x) - len(y)
        if len_diff > 0:
            y = (1,) * len_diff + y
        elif len_diff < 0:
            x = (1,) * (-len_diff) + x

        assert len(x) == len(y)

        out = []
        for idx, (xx, yy) in enumerate(zip(x, y)):
            if xx == 1:
                out.append(yy)
            elif yy == 1:
                out.append(xx)
            elif xx != yy:
                raise ValueError(f"Shape mismatch in dim {idx}: {xx} vs {yy}")
            else:
                out.append(xx)

        return tuple(out)


class DataExpr(Expr):
    def __str__(self) -> str:
        return "$"

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, DataExpr)

    @property
    def variables(self) -> Set[str]:
        return set()

    def eval(self, values: Dict[str, ShapeVariable]) -> ShapeVariable:
        raise RuntimeError("Tried to evaluate DataExpr")



from .checks.data import check_data
from .checks.func import check_func, check_func2
from .checks.shapes import check_shapes
from .contexts import disable_checks, enable_checks
from .parser.parse_cache import (
    parser_cache_clear,
    parser_cache_info,
    parser_resize_cache,
)

__all__ = [
    "check_data",
    "check_func",
    "check_func2",
    "check_shapes",
    "disable_checks",
    "enable_checks",
    "parser_cache_clear",
    "parser_cache_info",
    "parser_resize_cache",
]



from functools import _CacheInfo, lru_cache
from typing import Callable, Generic, Optional, TypeVar

from typing_extensions import ParamSpec

_T = TypeVar("_T")
_P = ParamSpec("_P")


class ResizeableLruCache(Generic[_P, _T]):
    def __init__(self, func: Callable[_P, _T], maxsize: Optional[int] = 128):
        self._func = func
        self._cached_func = lru_cache(maxsize=maxsize)(func)

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _T:
        return self._cached_func(*args, **kwargs)  # type: ignore[arg-type]

    def reset_maxsize(self, maxsize: Optional[int]) -> None:
        self._cached_func = lru_cache(maxsize=maxsize)(self._func)

    def cache_info(self) -> _CacheInfo:
        return self._cached_func.cache_info()

    def cache_clear(self) -> None:
        self._cached_func.cache_clear()


def resizeable_lru_cache(
    maxsize: Optional[int] = 128,
) -> Callable[[Callable[_P, _T]], ResizeableLruCache[_P, _T]]:
    """Resizable version of functools.lru_cache."""
    return lambda f: ResizeableLruCache(f, maxsize=maxsize)



import functools
import inspect
from typing import Any, Callable, Dict, Iterable, List, Mapping, TypeVar, Union

from eincheck.checks.shapes import check_shapes
from eincheck.contexts import _should_do_checks
from eincheck.parser.grammar import ShapeArg, create_shape_spec
from eincheck.parser.shape_spec import ShapeSpec
from eincheck.utils import get_object, parse_dot_name

_T_Callable = TypeVar("_T_Callable", bound=Callable[..., Any])


def check_func(
    shapes: str = "", **kwargs: ShapeArg
) -> Callable[[_T_Callable], _T_Callable]:
    """Check the input and output shapes of a function.


    :param shapes: string of input and output shape specs
    :param kwargs: additional shape specs for function inputs
    :return: a function decorator
    """
    if "->" in shapes:
        input_str, output_str = shapes.split("->", 2)
    else:
        input_str = ""
        output_str = shapes

    input_arg_shapes = [
        create_shape_spec(s.strip()) for s in input_str.split(",") if s.strip()
    ]
    input_kwarg_shapes = {k: create_shape_spec(v) for k, v in kwargs.items()}

    output_shapes = [
        create_shape_spec(s.strip()) for s in output_str.split(",") if s.strip()
    ]

    return _get_wrapper(input_arg_shapes, input_kwarg_shapes, output_shapes)


def check_func2(
    input_shapes: Union[str, Mapping[str, ShapeArg]],
    output_shapes: Union[str, Mapping[str, ShapeArg]] = "",
) -> Callable[[_T_Callable], _T_Callable]:
    """Check the input and output shapes of a function.

    This function is an alternative to ``check_func`` that works better with
    dictionaries.
    It takes an input spec and an output spec, where each spec can be either a
    dictionary or a comma separated string.

    If both are strings, ``check_func2(input_str, output_str)`` is equivalent to
    ``check_func(f"{input_str} -> {output_str}")``.

    If the input spec is a dictionary and the output spec is a string,
    ``check_func2(input_dict, output_str)`` is equivalent to
    ``check_func(output_str, **input_dict)``.

    This decorator also supports a dictionary for output shapes, which ``check_func``
    does not.
    This enables dotpath names on the returned object.

    Examples:

    .. doctest::

        >>> from eincheck import check_func2
        >>> from numpy.random import randn
        >>> from numpy.typing import NDArray
        >>> from typing import NamedTuple, Tuple
        >>>
        >>> Array = NDArray[float]
        >>>
        >>> # Three equivalent ways of using check_func2.
        >>> @check_func2("i, j", "i j, i j")
        ... def foo1(x: Array, y: Array) -> Tuple[Array, Array]:
        ...     return x[:, None] + y, x[:, None] * y
        ...
        >>> _ = foo1(randn(4), randn(5))
        >>>
        >>> @check_func2("i, j -> i j, i j")
        ... def foo2(x: Array, y: Array) -> Tuple[Array, Array]:
        ...     return x[:, None] + y, x[:, None] * y
        ...
        >>> _ = foo2(randn(4), randn(5))
        >>>
        >>> @check_func2({"x": "i", "y": "j"}, {"0": "i j", "1": "i j"})
        ... def foo3(x: Array, y: Array) -> Tuple[Array, Array]:
        ...     return x[:, None] + y, x[:, None] * y
        ...
        >>> _ = foo3(randn(4), randn(5))
        >>>
        >>> class Pair(NamedTuple):
        ...     first: Array
        ...     second: Array
        ...
        >>> @check_func2(
        ...     {"x.first": "*a", "x.second": "*b", "y.first": "*a", "y.second": "*b"},
        ...     {"first": "*a", "second": "*b"},
        ... )
        ... def add_pairs(x: Pair, y: Pair) -> Pair:
        ...     return Pair(x.first + y.first, x.second + y.second)
        ...
        >>> _ = add_pairs(Pair(randn(4), randn(5, 6)), Pair(randn(4), randn(5, 6)))

    :param input_shapes: comma separated string or dictionary of shapes
    :param output_shapes: comma separated string or dictionary of shapes
    :return: a function decorator
    """
    if isinstance(input_shapes, str) and "->" in input_shapes:
        if output_shapes:
            raise ValueError(
                "'->' in input_shapes should only be used when output_shapes is empty"
            )

        input_shapes, output_shapes = input_shapes.split("->", 2)

    if isinstance(input_shapes, str):
        input_arg_shapes = [
            create_shape_spec(s.strip()) for s in input_shapes.split(",") if s.strip()
        ]
        input_kwarg_shapes = {}
    else:
        input_arg_shapes = []
        input_kwarg_shapes = {k: create_shape_spec(v) for k, v in input_shapes.items()}

    parsed_output_shapes: Union[List[ShapeSpec], Dict[str, ShapeSpec]]
    if isinstance(output_shapes, str):
        parsed_output_shapes = [
            create_shape_spec(s.strip()) for s in output_shapes.split(",") if s.strip()
        ]
    else:
        parsed_output_shapes = {
            k: create_shape_spec(v) for k, v in output_shapes.items()
        }

    return _get_wrapper(input_arg_shapes, input_kwarg_shapes, parsed_output_shapes)


def _get_wrapper(
    input_arg_shapes: List[ShapeSpec],
    input_kwarg_shapes: Dict[str, ShapeSpec],
    output_shapes: Union[List[ShapeSpec], Dict[str, ShapeSpec]],
) -> Callable[[_T_Callable], _T_Callable]:
    def wrapper(func: _T_Callable) -> _T_Callable:
        input_shapes = input_kwarg_shapes

        sig = inspect.signature(func)
        if len(sig.parameters) < len(input_arg_shapes):
            raise ValueError(
                f"Expected at least {len(input_arg_shapes)} input parameters, "
                f"got {len(sig.parameters)}"
            )

        for arg_spec, arg_name in zip(input_arg_shapes, sig.parameters):
            if arg_name in input_shapes:
                raise ValueError(
                    f"Spec for {arg_name} specified in both args and kwargs."
                )

            input_shapes[arg_name] = arg_spec

        # dot_name, name_base, name_parts
        # e.g. ("foo.x.y", "foo", ["x", "y"])
        parsed_names = [(n, *parse_dot_name(n)) for n in input_shapes]
        # sort to match signature for nice error messages
        sig_params = list(sig.parameters)
        parsed_names.sort(
            key=lambda t: (
                sig_params.index(t[1]) if t[1] in sig_params else len(sig_params)
            )
        )
        _check_no_extra_params((x for _, x, _ in parsed_names), sig)

        @functools.wraps(func)
        def inner(*args: Any, **kwargs: Any) -> Any:
            if not _should_do_checks():
                return func(*args, **kwargs)

            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            input_data = {}
            for spec_name, spec_base, spec_parts in parsed_names:
                if spec_base not in sig.parameters:
                    continue

                p = sig.parameters[spec_base]
                x = get_object(spec_name, bound_args.arguments)

                if len(spec_parts) > 0:
                    p_data = [(spec_name, x)]
                elif p.kind is inspect.Parameter.VAR_POSITIONAL:
                    assert isinstance(x, tuple)
                    p_data = [
                        (f"{spec_name}_{x_idx}", xx) for x_idx, xx in enumerate(x)
                    ]
                elif p.kind is inspect.Parameter.VAR_KEYWORD:
                    assert isinstance(x, dict)
                    p_data = list(x.items())
                else:
                    p_data = [(spec_name, x)]

                for x_name, xx in p_data:
                    assert x_name not in input_data
                    input_data[x_name] = (xx, input_shapes[spec_name])

            updated_spec = check_shapes(**input_data)

            out = func(*args, **kwargs)

            if isinstance(output_shapes, list):
                out_tup = (
                    out if isinstance(out, tuple) and len(output_shapes) > 1 else (out,)
                )
                if len(out_tup) < len(output_shapes):
                    raise ValueError(
                        f"Expected at least {len(output_shapes)} outputs, "
                        f"got {len(out_tup)}"
                    )
                check_shapes(
                    **{
                        f"output{i}": (x, s)
                        for i, (x, s) in enumerate(zip(out_tup, output_shapes))
                    },
                    **updated_spec,
                )
            else:
                output_data = {
                    f"output {k}": (get_object(k, out), v)
                    for k, v in output_shapes.items()
                }
                check_shapes(**output_data, **updated_spec)

            return out

        return inner  # type: ignore[return-value]

    return wrapper


def _check_no_extra_params(got_names: Iterable[str], sig: inspect.Signature) -> None:
    """Check that all of got_names are valid for sig."""
    if any(p.kind is inspect.Parameter.VAR_KEYWORD for p in sig.parameters.values()):
        # Any param name is valid.
        return

    extra_names = set(got_names) - set(sig.parameters)
    if extra_names:
        raise NameError(
            f"Parameter names not found in function signature: {extra_names}"
        )



import functools
import itertools
import sys
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Mapping, Optional, Set, Tuple, TypeVar

from eincheck.checks.shapes import check_shapes
from eincheck.parser.grammar import ShapeArg, create_shape_spec
from eincheck.parser.shape_spec import ShapeSpec
from eincheck.utils import get_object, parse_dot_name

_T = TypeVar("_T")


class DataWrapper(ABC):
    module_name: Optional[str] = None

    @classmethod
    def can_load(cls) -> bool:
        return cls.module_name is None or cls.module_name in sys.modules

    @abstractmethod
    def is_match(self, x: Any) -> bool:
        """Whether x is a data object of the right type."""

    @abstractmethod
    def wrap(self, cls: _T, shapes: Mapping[str, ShapeSpec]) -> _T:
        pass

    @staticmethod
    def check_fields(shapes: Mapping[str, ShapeSpec], got: Set[str]) -> None:
        extra_names = {parse_dot_name(n)[0] for n in shapes} - got
        if extra_names:
            raise ValueError("No field found: [" + " ".join(sorted(extra_names)) + "]")

    @staticmethod
    def get_shapes_func(
        shapes: Mapping[str, ShapeSpec]
    ) -> Callable[[Any], Dict[str, Tuple[Any, ShapeSpec]]]:
        def get_shapes(self: Any) -> Dict[str, Tuple[Any, ShapeSpec]]:
            return {k: (get_object(k, self), s) for k, s in shapes.items()}

        return get_shapes


class NamedTupleWrapper(DataWrapper):
    def is_match(self, x: Any) -> bool:
        return issubclass(x, tuple) and hasattr(x, "_fields")

    def wrap(self, cls: _T, shapes: Mapping[str, ShapeSpec]) -> _T:
        self.check_fields(shapes, set(cls._fields))  # type: ignore[attr-defined]

        _new = cls.__new__

        cls._get_shapes = DataWrapper.get_shapes_func(  # type: ignore[attr-defined]
            shapes
        )

        @functools.wraps(_new)
        def new_new(*a: Any, **k: Any) -> Any:
            out = _new(*a, **k)
            check_shapes(**out._get_shapes())
            return out

        cls.__new__ = new_new  # type: ignore[method-assign]

        return cls


def _func_with_check(
    cls: Any, func: str, shapes: Mapping[str, ShapeSpec], append: bool
) -> None:
    old_f = getattr(cls, func)

    cls._get_shapes = DataWrapper.get_shapes_func(shapes)

    if append:

        def new_f(self: Any, *a: Any, **k: Any) -> Any:
            old_f(self, *a, **k)
            check_shapes(**self._get_shapes())

    else:

        def new_f(self: Any, *a: Any, **k: Any) -> Any:
            check_shapes(**self._get_shapes())
            old_f(self, *a, **k)

    new_f = functools.wraps(old_f)(new_f)
    setattr(cls, func, new_f)


class DataclassWrapper(DataWrapper):
    if sys.version_info[:2] < (3, 8):
        module_name = "dataclasses"

    def __init__(self) -> None:
        super().__init__()
        import dataclasses

        self.dataclasses = dataclasses

    def is_match(self, x: Any) -> bool:
        out = self.dataclasses.is_dataclass(x)
        assert isinstance(out, bool)
        return out

    def wrap(self, cls: _T, shapes: Mapping[str, ShapeSpec]) -> _T:
        self.check_fields(
            shapes,
            {f.name for f in self.dataclasses.fields(cls)},  # type: ignore[arg-type]
        )

        if hasattr(cls, "__post_init__"):
            _func_with_check(cls, "__post_init__", shapes, False)
        else:
            _func_with_check(cls, "__init__", shapes, True)
        return cls


class AttrsWrapper(DataWrapper):
    module_name = "attrs"

    def __init__(self) -> None:
        super().__init__()
        import attrs

        self.attrs = attrs

    def is_match(self, x: Any) -> bool:
        return self.attrs.has(x)

    def wrap(self, cls: _T, shapes: Mapping[str, ShapeSpec]) -> _T:
        self.check_fields(
            shapes, {a.name for a in self.attrs.fields(cls)}  # type: ignore[arg-type]
        )

        if hasattr(cls, "__attrs_post_init__"):
            _func_with_check(cls, "__attrs_post_init__", shapes, False)

        else:
            _func_with_check(cls, "__init__", shapes, True)

        return cls


_wrappers: List[DataWrapper] = []

_T_Data = TypeVar("_T_Data")


def check_data(
    shape_dict: Optional[Mapping[str, ShapeArg]] = None, /, **kwargs: ShapeArg
) -> Callable[[_T_Data], _T_Data]:
    """Check the shapes of fields of a data object.

    The currently supported data objects are NamedTuple, dataclasses, and attrs.

    :param shape_dict: shape specs for fields of the data object in a dictionary
    :param kwargs: shape specs for fields of the data object as keywords
    :return: a decorator for the data object class
    """
    if shape_dict is None:
        shape_dict = {}
    shapes = {
        k: create_shape_spec(v)
        for k, v in itertools.chain(shape_dict.items(), kwargs.items())
    }

    def wrapper(cls: _T_Data) -> _T_Data:
        for w in _wrappers:
            if w.is_match(cls):
                return w.wrap(cls, shapes)

        for w_cls in DataWrapper.__subclasses__():
            if w_cls.can_load():
                _wrappers.append(w_cls())  # type: ignore[abstract]
                if _wrappers[-1].is_match(cls):
                    return _wrappers[-1].wrap(cls, shapes)

        raise TypeError(f"Unexpected data type {cls}")

    return wrapper



from __future__ import annotations

import dataclasses
from typing import Dict, List, Optional, Set, Tuple

from eincheck.parser.dim_spec import DimSpec
from eincheck.parser.expressions import DataExpr
from eincheck.types import ShapeVariable


@dataclasses.dataclass
class ShapeSpec:
    dims: List[DimSpec]
    _unknown_n_dims_indices_cache: Optional[Tuple[int, List[int]]] = dataclasses.field(
        default=None
    )

    def __str__(self) -> str:
        return "[" + " ".join(str(d) for d in self.dims) + "]"

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, ShapeSpec) and __o.dims == self.dims

    def is_checkable(self, bindings: Dict[str, ShapeVariable]) -> bool:
        if len(self.unknown_n_dims_indices(bindings)) > 1:
            return False

        return all(d.is_checkable(bindings) for d in self.dims)

    def unknown_n_dims_indices(self, bindings: Dict[str, ShapeVariable]) -> List[int]:
        key = hash(tuple(bindings.items()))
        if (
            self._unknown_n_dims_indices_cache
            and self._unknown_n_dims_indices_cache[0] == key
        ):
            return self._unknown_n_dims_indices_cache[1]

        out = [i for i, d in enumerate(self.dims) if d.n_dims(bindings) is None]
        self._unknown_n_dims_indices_cache = (key, out)
        return out

    def min_rank(self, bindings: Dict[str, ShapeVariable]) -> int:
        n_dims = [d.n_dims(bindings) for d in self.dims]
        return sum(x for x in n_dims if x is not None)

    def matched_indices(
        self, bindings: Dict[str, ShapeVariable], n_dims: int
    ) -> List[Tuple[DimSpec, int, int]]:
        n_dims_per_spec = [d.n_dims(bindings) for d in self.dims]
        min_rank = sum(x for x in n_dims_per_spec if x is not None)
        if min_rank > n_dims:
            raise RuntimeError(f"Expected rank at least {min_rank}, got {n_dims}")

        none_inds = [i for i, x in enumerate(n_dims_per_spec) if x is None]
        if len(none_inds) > 1:
            raise RuntimeError(
                "Cannot determine matched indices: "
                f"multiple DimSpec with unknown n_dims {none_inds}"
            )

        i = 0
        output = []
        for d, x in zip(self.dims, n_dims_per_spec):
            y = (n_dims - min_rank) if x is None else x
            output.append((d, i, i + y))
            i += y

        if i != n_dims:
            raise RuntimeError(f"Expected rank {i}, got {n_dims}")

        return output

    @property
    def is_data_expr(self) -> bool:
        return len(self.dims) == 1 and isinstance(self.dims[0].value, DataExpr)

    @property
    def variables(self) -> Set[str]:
        x = set()
        for d in self.dims:
            if d.value is not None:
                x |= d.value.variables
        return x
